@@ -1,7 +1,6 @@
 /* minimal version of PG 9.6's float8out_internal function for use in 9.4 and 9.5 */
 
 #define MAXDOUBLEWIDTH	128
-#define DBL_DIG			15
 extern int		extra_float_digits;
 
 static char *
@@ -12,6 +11,12 @@ float8out_internal(double num)
 
 	if (isnan(num))
 		return strcpy(ascii, "NaN");
+	if (!isfinite(num)) {
+		if (num > 0)
+			return strcpy(ascii, "Infinity");
+		else
+			return strcpy(ascii, "-Infinity");
+	}
 
 	if (ndig < 1)
 		ndig = 1;
