@@ -14,8 +14,12 @@ GNU General Public License for more details.
 
 #include "postgres.h"
 #include "fmgr.h"
+#include "libpq/pqformat.h" /* send/recv */
 #include "utils/builtins.h" /* cstring_to_text (needed on 9.5) */
 #include "utils/guc.h"
+#if PG_VERSION_NUM >= 120000
+#include "utils/float.h" /* extra_float_digits */
+#endif
 #include "utils/hsearch.h"
 #include <math.h>
 
@@ -32,6 +36,7 @@ GNU General Public License for more details.
 /* global variables */
 
 static bool unit_output_superscript;
+static bool unit_time_output_custom;
 static bool unit_byte_output_iec;
 static bool unit_output_base_units;
 
@@ -149,6 +154,19 @@ _PG_init(void)
 			NULL,
 			NULL);
 
+	DefineCustomBoolVariable("unit.time_output_custom",
+			"Output time values in minutes, hours, days, and julianyears",
+			"Set to on to output time unit values greater or equal to 60 s "
+			"using minutes, hours, days, and julianyears instead of seconds "
+			"with SI prefixes",
+			&unit_time_output_custom,
+			true,
+			PGC_USERSET,
+			0, /* no flags */
+			NULL,
+			NULL,
+			NULL);
+
 	DefineCustomBoolVariable("unit.byte_output_iec",
 			"Output byte unit values with binary IEC prefixes",
 			"Set to on to output byte unit values using binary IEC prefixes "
@@ -214,6 +232,71 @@ print_exponent (char **output_p, int e)
 	}
 }
 
+/* this function is used to format time values of at least one minute */
+static void
+print_time_interval (char **output_p, double t)
+{
+	int		 h, m;
+	char	*sign = "+";
+	int		 ndig;
+
+	/* print - */
+	if (t < 0) {
+		*output_p += sprintf(*output_p, "-");
+		t = -t;
+		sign = "-";
+	}
+
+	ndig = DBL_DIG + extra_float_digits - log10(t); /* adjust seconds precision for d/h/min already printed */
+
+	/* print years */
+	if (t >= TIME_YEAR) {
+		double years = trunc(t / TIME_YEAR);
+		t = fmod(t, TIME_YEAR);
+		*output_p += sprintf(*output_p, "%g " TIME_YEAR_NAME, years);
+		if (t != 0)
+			*output_p += sprintf(*output_p, " %s ", sign);
+	}
+	if (t == 0)
+		return;
+
+	/* print days */
+	if (t >= TIME_DAY) {
+		int days = trunc(t / TIME_DAY);
+		t = fmod(t, TIME_DAY);
+		*output_p += sprintf(*output_p, "%d d", days);
+		if (t != 0)
+			*output_p += sprintf(*output_p, " %s ", sign);
+	}
+	if (t == 0)
+		return;
+
+	/* print hh:mm: */
+	h = trunc(t / TIME_HOUR);
+	t = fmod(t, TIME_HOUR);
+	m = trunc(t / TIME_MINUTE);
+	t = fmod(t, TIME_MINUTE);
+	*output_p += sprintf(*output_p, "%02d:%02d:", h, m);
+
+	/* print ss.sss */
+	if (t < 10.0) /* zero-pad */
+		*output_p += sprintf(*output_p, "0");
+	if (ndig < 0)
+		ndig = 0;
+	*output_p += sprintf(*output_p, "%.*f", ndig, t);
+	if (ndig > 0) {
+		/* strip trailing zeros from %f format */
+		while (*(*output_p - 1) == '0') {
+			--*output_p;
+		}
+		if (*(*output_p - 1) == '.') {
+			--*output_p;
+		}
+		**output_p = '\0';
+	}
+	*output_p += sprintf(*output_p, " s");
+}
+
 /* format Unit as string */
 char *
 unit_cstring (Unit *unit)
@@ -327,6 +410,15 @@ unit_cstring (Unit *unit)
 		double	 v_abs = fabs(unit->value);
 		char	*prefix = "";
 		double	 factor = 1.0;
+		char	*unit_name = derived_unit ? derived_unit->name : (char *)base_units[u_numerator];
+
+		/* case 2a: minute/hour/day output requested, unit is seconds, and
+		 * absolute value is >= 60 seconds. Print as interval and exit */
+		if (unit_time_output_custom && derived_unit &&
+				!strcmp(derived_unit->name, "s") && v_abs >= TIME_MINUTE) {
+			print_time_interval(&output_p, unit->value);
+			return output;
+		}
 
 		if (v_abs >= 1e27) {
 			// do nothing
@@ -342,6 +434,8 @@ unit_cstring (Unit *unit)
 			prefix = "T"; factor = 1e-12;
 		} else if (v_abs >= POWER_9) {
 			prefix = "G"; factor = 1e-9;
+			if (!strcmp(unit_name, "s"))
+				unit_name = "sec"; /* avoid Gs because it's gauss, not time */
 		} else if (v_abs >= POWER_6) {
 			prefix = "M"; factor = 1e-6;
 		} else if (v_abs >= POWER_3) {
@@ -366,16 +460,15 @@ unit_cstring (Unit *unit)
 			prefix = "y"; factor = 1e24;
 		} /* else do nothing */
 
-		/* case 2a: derived unit, print with SI prefix and exit */
-		if (derived_unit) {
-			print_output("%s %s%s", float8out_internal (unit->value * factor),
-					prefix, derived_unit->name);
-			return output;
-		}
-
-		/* case 2b: single unit in numerator (exponent 1), print with SI prefix and continue */
+		/* print with SI prefix */
 		print_output("%s %s%s", float8out_internal (unit->value * factor),
-				prefix, base_units[u_numerator]);
+				prefix, unit_name);
+
+		/* case 2b: derived unit: stop here */
+		if (derived_unit)
+			return output;
+
+		/* case 2c: single unit in numerator (exponent 1): continue */
 		numerator = true;
 
 	/* case 3: zero or more than one unit in numerator */
@@ -450,6 +543,33 @@ unit_out(PG_FUNCTION_ARGS)
 	PG_RETURN_CSTRING(unit_cstring(unit));
 }
 
+PG_FUNCTION_INFO_V1(unit_recv);
+
+Datum
+unit_recv(PG_FUNCTION_ARGS)
+{
+	StringInfo	 buf = (StringInfo) PG_GETARG_POINTER(0);
+	Unit		*unit = palloc(sizeof(Unit));
+
+	unit->value = pq_getmsgfloat8(buf);
+	memcpy(unit->units, pq_getmsgbytes(buf, N_UNITS), N_UNITS);
+	PG_RETURN_POINTER(unit);
+}
+
+PG_FUNCTION_INFO_V1(unit_send);
+
+Datum
+unit_send(PG_FUNCTION_ARGS)
+{
+	Unit		*unit = (Unit *) PG_GETARG_POINTER(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+	pq_sendfloat8(&buf, unit->value);
+	pq_sendbytes(&buf, (const char *)unit->units, N_UNITS);
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
 /* constructors */
 
 PG_FUNCTION_INFO_V1 (dbl2unit);
@@ -608,6 +728,19 @@ unit_round(PG_FUNCTION_ARGS)
 	PG_RETURN_POINTER(result);
 }
 
+PG_FUNCTION_INFO_V1(unit_diff);
+
+Datum
+unit_diff(PG_FUNCTION_ARGS)
+{
+	Unit	*a = (Unit *) PG_GETARG_POINTER(0);
+	Unit	*b = (Unit *) PG_GETARG_POINTER(1);
+	Unit	 result;
+
+	unit_sub_internal(a, b, &result);
+	PG_RETURN_FLOAT8(result.value);
+}
+
 /* operators */
 
 PG_FUNCTION_INFO_V1(unit_add);
@@ -855,6 +988,7 @@ unit_at(PG_FUNCTION_ARGS)
 				b));
 }
 
+/* obsolete version of unit_at_text2 used in v4..6 */
 /* needs search_path = @extschema@ due to use of unit_parse() */
 PG_FUNCTION_INFO_V1(unit_at_text);
 
@@ -881,6 +1015,32 @@ unit_at_text(PG_FUNCTION_ARGS)
 				b));
 }
 
+/* needs search_path = @extschema@ due to use of unit_parse() */
+PG_FUNCTION_INFO_V1(unit_at_text2);
+
+Datum
+unit_at_text2(PG_FUNCTION_ARGS)
+{
+	Unit		*a = (Unit *) PG_GETARG_POINTER(0);
+	char		*b = text_to_cstring(PG_GETARG_TEXT_PP(1));
+	UnitShift	 bu;
+
+	if (unit_parse(b, &bu) > 0)
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
+				 errmsg("invalid input syntax for unit: \"%s\", %s",
+					 b, yyerrstr)));
+	test_same_dimension("@", a, &bu.unit);
+	if (bu.unit.value == 0)
+		ereport(ERROR,
+				(errcode(ERRCODE_DIVISION_BY_ZERO),
+				 errmsg("division by zero-valued unit: \"%s\"", b)));
+	PG_RETURN_TEXT_P(cstring_to_text(psprintf("%s %s%s",
+				float8out_internal ((a->value - bu.shift) / bu.unit.value),
+				(atof(b) > 0 ? "* " : ""),
+				b)));
+}
+
 /* needs search_path = @extschema@ due to use of unit_parse() */
 PG_FUNCTION_INFO_V1(unit_at_double);
 
@@ -993,6 +1153,96 @@ unit_cmp(PG_FUNCTION_ARGS)
 	PG_RETURN_INT32(unit_cmp_internal(a, b));
 }
 
+/* strict comparisons (error out when dimensions do not match) */
+
+static int
+unit_strict_cmp_internal(Unit *a, Unit *b)
+{
+	test_same_dimension("strict comparison", a, b);
+	if (a->value < b->value)
+		return -1;
+	if (a->value > b->value)
+		return 1;
+	return memcmp(a->units, b->units, N_UNITS);
+}
+
+PG_FUNCTION_INFO_V1(unit_strict_lt);
+
+Datum
+unit_strict_lt(PG_FUNCTION_ARGS)
+{
+	Unit	*a = (Unit *) PG_GETARG_POINTER(0);
+	Unit	*b = (Unit *) PG_GETARG_POINTER(1);
+
+	PG_RETURN_BOOL(unit_strict_cmp_internal(a, b) < 0);
+}
+
+PG_FUNCTION_INFO_V1(unit_strict_le);
+
+Datum
+unit_strict_le(PG_FUNCTION_ARGS)
+{
+	Unit	*a = (Unit *) PG_GETARG_POINTER(0);
+	Unit	*b = (Unit *) PG_GETARG_POINTER(1);
+
+	PG_RETURN_BOOL(unit_strict_cmp_internal(a, b) <= 0);
+}
+
+PG_FUNCTION_INFO_V1(unit_strict_eq);
+
+Datum
+unit_strict_eq(PG_FUNCTION_ARGS)
+{
+	Unit	*a = (Unit *) PG_GETARG_POINTER(0);
+	Unit	*b = (Unit *) PG_GETARG_POINTER(1);
+
+	PG_RETURN_BOOL(unit_strict_cmp_internal(a, b) == 0);
+}
+
+PG_FUNCTION_INFO_V1(unit_strict_ne);
+
+Datum
+unit_strict_ne(PG_FUNCTION_ARGS)
+{
+	Unit	*a = (Unit *) PG_GETARG_POINTER(0);
+	Unit	*b = (Unit *) PG_GETARG_POINTER(1);
+
+	PG_RETURN_BOOL(unit_strict_cmp_internal(a, b) != 0);
+}
+
+PG_FUNCTION_INFO_V1(unit_strict_ge);
+
+Datum
+unit_strict_ge(PG_FUNCTION_ARGS)
+{
+	Unit	*a = (Unit *) PG_GETARG_POINTER(0);
+	Unit	*b = (Unit *) PG_GETARG_POINTER(1);
+
+	PG_RETURN_BOOL(unit_strict_cmp_internal(a, b) >= 0);
+}
+
+PG_FUNCTION_INFO_V1(unit_strict_gt);
+
+Datum
+unit_strict_gt(PG_FUNCTION_ARGS)
+{
+	Unit	*a = (Unit *) PG_GETARG_POINTER(0);
+	Unit	*b = (Unit *) PG_GETARG_POINTER(1);
+
+	PG_RETURN_BOOL(unit_strict_cmp_internal(a, b) > 0);
+}
+
+PG_FUNCTION_INFO_V1(unit_strict_cmp);
+
+Datum
+unit_strict_cmp(PG_FUNCTION_ARGS)
+{
+	Unit	*a = (Unit *) PG_GETARG_POINTER(0);
+	Unit	*b = (Unit *) PG_GETARG_POINTER(1);
+
+	PG_RETURN_INT32(unit_strict_cmp_internal(a, b));
+}
+
 PG_FUNCTION_INFO_V1(unit_least);
 
 Datum
